@@ -21,21 +21,17 @@
 
 import datetime
 
+import cryptography.hazmat.primitives.asymmetric.ec
 import cryptography.hazmat.primitives.asymmetric.rsa
 import cryptography.x509
 from dateutil.relativedelta import relativedelta
 
 import gimmecert.crypto
 
+import pytest
 from freezegun import freeze_time
 
 
-def test_generate_private_key_returns_private_key():
-    private_key = gimmecert.crypto.generate_private_key()
-
-    assert isinstance(private_key, cryptography.hazmat.primitives.asymmetric.rsa.RSAPrivateKey)
-
-
 def test_get_dn():
     dn = gimmecert.crypto.get_dn('My test')
     assert isinstance(dn, cryptography.x509.Name)
@@ -79,8 +75,8 @@ def test_issue_certificate_returns_certificate():
 
     issuer_dn = gimmecert.crypto.get_dn('My test 1')
     subject_dn = gimmecert.crypto.get_dn('My test 2')
-    issuer_private_key = gimmecert.crypto.generate_private_key()
-    subject_private_key = gimmecert.crypto.generate_private_key()
+    issuer_private_key = gimmecert.crypto.KeyGenerator('rsa', 2048)()
+    subject_private_key = gimmecert.crypto.KeyGenerator('rsa', 2048)()
     not_before, not_after = gimmecert.crypto.get_validity_range()
 
     certificate = gimmecert.crypto.issue_certificate(issuer_dn, subject_dn, issuer_private_key, subject_private_key.public_key(), not_before, not_after)
@@ -91,8 +87,8 @@ def test_issue_certificate_returns_certificate():
 def test_issue_certificate_has_correct_content():
     issuer_dn = gimmecert.crypto.get_dn('My test 1')
     subject_dn = gimmecert.crypto.get_dn('My test 2')
-    issuer_private_key = gimmecert.crypto.generate_private_key()
-    subject_private_key = gimmecert.crypto.generate_private_key()
+    issuer_private_key = gimmecert.crypto.KeyGenerator('rsa', 2048)()
+    subject_private_key = gimmecert.crypto.KeyGenerator('rsa', 2048)()
     not_before, not_after = gimmecert.crypto.get_validity_range()
 
     certificate = gimmecert.crypto.issue_certificate(issuer_dn, subject_dn, issuer_private_key, subject_private_key.public_key(), not_before, not_after)
@@ -107,7 +103,7 @@ def test_generate_ca_hierarchy_returns_list_with_3_elements_for_depth_3():
     base_name = 'My Project'
     depth = 3
 
-    hierarchy = gimmecert.crypto.generate_ca_hierarchy(base_name, depth)
+    hierarchy = gimmecert.crypto.generate_ca_hierarchy(base_name, depth, gimmecert.crypto.KeyGenerator("rsa", 2048))
 
     assert isinstance(hierarchy, list)
     assert len(hierarchy) == depth
@@ -117,28 +113,33 @@ def test_generate_ca_hierarchy_returns_list_with_1_element_for_depth_1():
     base_name = 'My Project'
     depth = 1
 
-    hierarchy = gimmecert.crypto.generate_ca_hierarchy(base_name, depth)
+    hierarchy = gimmecert.crypto.generate_ca_hierarchy(base_name, depth, gimmecert.crypto.KeyGenerator("rsa", 2048))
 
     assert isinstance(hierarchy, list)
     assert len(hierarchy) == depth
 
 
-def test_generate_ca_hierarchy_returns_list_of_private_key_certificate_pairs():
+@pytest.mark.parametrize("key_specification, private_key_instance_type", [
+    [("rsa", 1024), cryptography.hazmat.primitives.asymmetric.rsa.RSAPrivateKey],
+    [("ecdsa", cryptography.hazmat.primitives.asymmetric.ec.SECP192R1), cryptography.hazmat.primitives.asymmetric.ec.EllipticCurvePrivateKey],
+])
+def test_generate_ca_hierarchy_returns_list_of_private_key_certificate_pairs(key_specification, private_key_instance_type):
     base_name = 'My Project'
     depth = 3
 
-    hierarchy = gimmecert.crypto.generate_ca_hierarchy(base_name, depth)
+    hierarchy = gimmecert.crypto.generate_ca_hierarchy(base_name, depth, gimmecert.crypto.KeyGenerator(*key_specification))
 
     for private_key, certificate in hierarchy:
-        assert isinstance(private_key, cryptography.hazmat.primitives.asymmetric.rsa.RSAPrivateKey)
+        assert isinstance(private_key, private_key_instance_type)
         assert isinstance(certificate, cryptography.x509.Certificate)
 
 
 def test_generate_ca_hierarchy_subject_dns_have_correct_value():
     base_name = 'My Project'
     depth = 3
+    key_generator = gimmecert.crypto.KeyGenerator("rsa", 2048)
 
-    level1, level2, level3 = [certificate for _, certificate in gimmecert.crypto.generate_ca_hierarchy(base_name, depth)]
+    level1, level2, level3 = [certificate for _, certificate in gimmecert.crypto.generate_ca_hierarchy(base_name, depth, key_generator)]
 
     assert level1.subject == cryptography.x509.Name(gimmecert.crypto.get_dn('My Project Level 1 CA'))
     assert level2.subject == cryptography.x509.Name(gimmecert.crypto.get_dn('My Project Level 2 CA'))
@@ -149,7 +150,7 @@ def test_generate_ca_hierarchy_issuer_dns_have_correct_value():
     base_name = 'My Project'
     depth = 3
 
-    hierarchy = gimmecert.crypto.generate_ca_hierarchy(base_name, depth)
+    hierarchy = gimmecert.crypto.generate_ca_hierarchy(base_name, depth, gimmecert.crypto.KeyGenerator("rsa", 2048))
 
     level1_key, level1_certificate = hierarchy[0]
     level2_key, level2_certificate = hierarchy[1]
@@ -160,11 +161,15 @@ def test_generate_ca_hierarchy_issuer_dns_have_correct_value():
     assert level3_certificate.issuer == cryptography.x509.Name(gimmecert.crypto.get_dn('My Project Level 2 CA'))
 
 
-def test_generate_ca_hierarchy_private_keys_match_with_public_keys_in_certificates():
+@pytest.mark.parametrize("key_specification", [
+    ("rsa", 2048),
+    ("ecdsa", cryptography.hazmat.primitives.asymmetric.ec.SECP192R1)
+])
+def test_generate_ca_hierarchy_private_keys_match_with_public_keys_in_certificates(key_specification):
     base_name = 'My Project'
     depth = 3
 
-    hierarchy = gimmecert.crypto.generate_ca_hierarchy(base_name, depth)
+    hierarchy = gimmecert.crypto.generate_ca_hierarchy(base_name, depth, gimmecert.crypto.KeyGenerator(*key_specification))
 
     level1_private_key, level1_certificate = hierarchy[0]
     level2_private_key, level2_certificate = hierarchy[1]
@@ -175,11 +180,15 @@ def test_generate_ca_hierarchy_private_keys_match_with_public_keys_in_certificat
     assert level3_private_key.public_key().public_numbers() == level3_certificate.public_key().public_numbers()
 
 
-def test_generate_ca_hierarchy_cas_have_differing_keys():
+@pytest.mark.parametrize("key_specification", [
+    ("rsa", 2048),
+    ("ecdsa", cryptography.hazmat.primitives.asymmetric.ec.SECP192R1)
+])
+def test_generate_ca_hierarchy_cas_have_differing_keys(key_specification):
     base_name = 'My Project'
     depth = 3
 
-    hierarchy = gimmecert.crypto.generate_ca_hierarchy(base_name, depth)
+    hierarchy = gimmecert.crypto.generate_ca_hierarchy(base_name, depth, gimmecert.crypto.KeyGenerator(*key_specification))
 
     level1_private_key, _ = hierarchy[0]
     level2_private_key, _ = hierarchy[1]
@@ -198,7 +207,7 @@ def test_generate_ca_hierarchy_certificates_have_same_validity():
     base_name = 'My Project'
     depth = 3
 
-    hierarchy = gimmecert.crypto.generate_ca_hierarchy(base_name, depth)
+    hierarchy = gimmecert.crypto.generate_ca_hierarchy(base_name, depth, gimmecert.crypto.KeyGenerator("rsa", 2048))
 
     _, level1_certificate = hierarchy[0]
     _, level2_certificate = hierarchy[1]
@@ -210,7 +219,7 @@ def test_generate_ca_hierarchy_certificates_have_same_validity():
 
 def test_issue_certificate_sets_extensions():
     dn = gimmecert.crypto.get_dn('My test 1')
-    private_key = gimmecert.crypto.generate_private_key()
+    private_key = gimmecert.crypto.KeyGenerator('rsa', 2048)()
     not_before, not_after = gimmecert.crypto.get_validity_range()
     basic_constraints = cryptography.x509.BasicConstraints(ca=True, path_length=None)
     ocsp_no_check = cryptography.x509.OCSPNoCheck()
@@ -234,7 +243,7 @@ def test_issue_certificate_sets_extensions():
 
 def test_issue_certificate_sets_no_extensions_if_none_are_passed():
     dn = gimmecert.crypto.get_dn('My test 1')
-    private_key = gimmecert.crypto.generate_private_key()
+    private_key = gimmecert.crypto.KeyGenerator('rsa', 2048)()
     not_before, not_after = gimmecert.crypto.get_validity_range()
 
     certificate1 = gimmecert.crypto.issue_certificate(dn, dn, private_key, private_key.public_key(), not_before, not_after, None)
@@ -248,7 +257,7 @@ def test_generate_ca_hierarchy_produces_certificates_with_ca_basic_constraints()
     base_name = 'My test'
     depth = 3
 
-    hierarchy = gimmecert.crypto.generate_ca_hierarchy(base_name, depth)
+    hierarchy = gimmecert.crypto.generate_ca_hierarchy(base_name, depth, gimmecert.crypto.KeyGenerator("rsa", 2048))
 
     for _, certificate in hierarchy:
         stored_extension = certificate.extensions.get_extension_for_class(cryptography.x509.BasicConstraints)
@@ -261,10 +270,10 @@ def test_generate_ca_hierarchy_produces_certificates_with_ca_basic_constraints()
 
 
 def test_issue_server_certificate_returns_certificate():
-    ca_hierarchy = gimmecert.crypto.generate_ca_hierarchy('My Project', 1)
+    ca_hierarchy = gimmecert.crypto.generate_ca_hierarchy('My Project', 1, gimmecert.crypto.KeyGenerator("rsa", 2048))
     issuer_private_key, issuer_certificate = ca_hierarchy[0]
 
-    private_key = gimmecert.crypto.generate_private_key()
+    private_key = gimmecert.crypto.KeyGenerator('rsa', 2048)()
 
     certificate = gimmecert.crypto.issue_server_certificate('myserver', private_key.public_key(), issuer_private_key, issuer_certificate)
 
@@ -272,10 +281,10 @@ def test_issue_server_certificate_returns_certificate():
 
 
 def test_issue_server_certificate_sets_correct_extensions():
-    ca_hierarchy = gimmecert.crypto.generate_ca_hierarchy('My Project', 1)
+    ca_hierarchy = gimmecert.crypto.generate_ca_hierarchy('My Project', 1, gimmecert.crypto.KeyGenerator("rsa", 2048))
     issuer_private_key, issuer_certificate = ca_hierarchy[0]
 
-    private_key = gimmecert.crypto.generate_private_key()
+    private_key = gimmecert.crypto.KeyGenerator('rsa', 2048)()
 
     expected_basic_constraints = cryptography.x509.BasicConstraints(ca=False, path_length=None)
     expected_key_usage = cryptography.x509.KeyUsage(
@@ -317,10 +326,10 @@ def test_issue_server_certificate_sets_correct_extensions():
 
 
 def test_issue_server_certificate_has_correct_issuer_and_subject():
-    ca_hierarchy = gimmecert.crypto.generate_ca_hierarchy('My Project', 4)
+    ca_hierarchy = gimmecert.crypto.generate_ca_hierarchy('My Project', 4, gimmecert.crypto.KeyGenerator("rsa", 2048))
     issuer_private_key, issuer_certificate = ca_hierarchy[3]
 
-    private_key = gimmecert.crypto.generate_private_key()
+    private_key = gimmecert.crypto.KeyGenerator('rsa', 2048)()
 
     certificate = gimmecert.crypto.issue_server_certificate('myserver', private_key.public_key(), issuer_private_key, issuer_certificate)
 
@@ -328,11 +337,15 @@ def test_issue_server_certificate_has_correct_issuer_and_subject():
     assert certificate.subject == gimmecert.crypto.get_dn('myserver')
 
 
-def test_issue_server_certificate_has_correct_public_key():
-    ca_hierarchy = gimmecert.crypto.generate_ca_hierarchy('My Project', 1)
+@pytest.mark.parametrize("key_specification", [
+    ("rsa", 2048),
+    ("ecdsa", cryptography.hazmat.primitives.asymmetric.ec.SECP192R1)
+])
+def test_issue_server_certificate_has_correct_public_key(key_specification):
+    ca_hierarchy = gimmecert.crypto.generate_ca_hierarchy('My Project', 1, gimmecert.crypto.KeyGenerator("rsa", 2048))
     issuer_private_key, issuer_certificate = ca_hierarchy[0]
 
-    private_key = gimmecert.crypto.generate_private_key()
+    private_key = gimmecert.crypto.KeyGenerator(*key_specification)()
 
     certificate = gimmecert.crypto.issue_server_certificate('myserver', private_key.public_key(), issuer_private_key, issuer_certificate)
 
@@ -341,10 +354,10 @@ def test_issue_server_certificate_has_correct_public_key():
 
 @freeze_time('2018-01-01 00:15:00')
 def test_issue_server_certificate_not_before_is_15_minutes_in_past():
-    ca_hierarchy = gimmecert.crypto.generate_ca_hierarchy('My Project', 1)
+    ca_hierarchy = gimmecert.crypto.generate_ca_hierarchy('My Project', 1, gimmecert.crypto.KeyGenerator("rsa", 2048))
     issuer_private_key, issuer_certificate = ca_hierarchy[0]
 
-    private_key = gimmecert.crypto.generate_private_key()
+    private_key = gimmecert.crypto.KeyGenerator('rsa', 2048)()
 
     certificate = gimmecert.crypto.issue_server_certificate('myserver', private_key.public_key(), issuer_private_key, issuer_certificate)
 
@@ -353,11 +366,11 @@ def test_issue_server_certificate_not_before_is_15_minutes_in_past():
 
 def test_issue_server_certificate_not_before_does_not_exceed_ca_validity():
     with freeze_time('2018-01-01 00:15:00'):
-        ca_hierarchy = gimmecert.crypto.generate_ca_hierarchy('My Project', 1)
+        ca_hierarchy = gimmecert.crypto.generate_ca_hierarchy('My Project', 1, gimmecert.crypto.KeyGenerator("rsa", 2048))
 
     issuer_private_key, issuer_certificate = ca_hierarchy[0]
 
-    private_key = gimmecert.crypto.generate_private_key()
+    private_key = gimmecert.crypto.KeyGenerator('rsa', 2048)()
 
     with freeze_time(issuer_certificate.not_valid_before - datetime.timedelta(seconds=1)):
         certificate1 = gimmecert.crypto.issue_server_certificate('myserver', private_key.public_key(), issuer_private_key, issuer_certificate)
@@ -367,11 +380,11 @@ def test_issue_server_certificate_not_before_does_not_exceed_ca_validity():
 
 def test_issue_server_certificate_not_after_does_not_exceed_ca_validity():
     with freeze_time('2018-01-01 00:15:00'):
-        ca_hierarchy = gimmecert.crypto.generate_ca_hierarchy('My Project', 1)
+        ca_hierarchy = gimmecert.crypto.generate_ca_hierarchy('My Project', 1, gimmecert.crypto.KeyGenerator("rsa", 2048))
 
     issuer_private_key, issuer_certificate = ca_hierarchy[0]
 
-    private_key = gimmecert.crypto.generate_private_key()
+    private_key = gimmecert.crypto.KeyGenerator('rsa', 2048)()
 
     with freeze_time(issuer_certificate.not_valid_after + datetime.timedelta(seconds=1)):
         certificate1 = gimmecert.crypto.issue_server_certificate('myserver', private_key.public_key(), issuer_private_key, issuer_certificate)
@@ -380,10 +393,10 @@ def test_issue_server_certificate_not_after_does_not_exceed_ca_validity():
 
 
 def test_issue_server_certificate_incorporates_additional_dns_subject_alternative_names():
-    ca_hierarchy = gimmecert.crypto.generate_ca_hierarchy('My Project', 1)
+    ca_hierarchy = gimmecert.crypto.generate_ca_hierarchy('My Project', 1, gimmecert.crypto.KeyGenerator("rsa", 2048))
     issuer_private_key, issuer_certificate = ca_hierarchy[0]
 
-    private_key = gimmecert.crypto.generate_private_key()
+    private_key = gimmecert.crypto.KeyGenerator('rsa', 2048)()
 
     expected_subject_alternative_name = cryptography.x509.SubjectAlternativeName(
         [
@@ -401,10 +414,10 @@ def test_issue_server_certificate_incorporates_additional_dns_subject_alternativ
 
 
 def test_issue_client_certificate_returns_certificate():
-    ca_hierarchy = gimmecert.crypto.generate_ca_hierarchy('My Project', 1)
+    ca_hierarchy = gimmecert.crypto.generate_ca_hierarchy('My Project', 1, gimmecert.crypto.KeyGenerator("rsa", 2048))
     issuer_private_key, issuer_certificate = ca_hierarchy[0]
 
-    private_key = gimmecert.crypto.generate_private_key()
+    private_key = gimmecert.crypto.KeyGenerator('rsa', 2048)()
 
     certificate = gimmecert.crypto.issue_client_certificate('myclient', private_key.public_key(), issuer_private_key, issuer_certificate)
 
@@ -412,10 +425,10 @@ def test_issue_client_certificate_returns_certificate():
 
 
 def test_issue_client_certificate_has_correct_issuer_and_subject():
-    ca_hierarchy = gimmecert.crypto.generate_ca_hierarchy('My Project', 4)
+    ca_hierarchy = gimmecert.crypto.generate_ca_hierarchy('My Project', 4, gimmecert.crypto.KeyGenerator("rsa", 2048))
     issuer_private_key, issuer_certificate = ca_hierarchy[3]
 
-    private_key = gimmecert.crypto.generate_private_key()
+    private_key = gimmecert.crypto.KeyGenerator('rsa', 2048)()
 
     certificate = gimmecert.crypto.issue_client_certificate('myclient', private_key.public_key(), issuer_private_key, issuer_certificate)
 
@@ -424,10 +437,10 @@ def test_issue_client_certificate_has_correct_issuer_and_subject():
 
 
 def test_issue_client_certificate_sets_correct_extensions():
-    ca_hierarchy = gimmecert.crypto.generate_ca_hierarchy('My Project', 1)
+    ca_hierarchy = gimmecert.crypto.generate_ca_hierarchy('My Project', 1, gimmecert.crypto.KeyGenerator("rsa", 2048))
     issuer_private_key, issuer_certificate = ca_hierarchy[0]
 
-    private_key = gimmecert.crypto.generate_private_key()
+    private_key = gimmecert.crypto.KeyGenerator('rsa', 2048)()
 
     expected_basic_constraints = cryptography.x509.BasicConstraints(ca=False, path_length=None)
     expected_key_usage = cryptography.x509.KeyUsage(
@@ -460,11 +473,15 @@ def test_issue_client_certificate_sets_correct_extensions():
     assert certificate.extensions.get_extension_for_class(cryptography.x509.ExtendedKeyUsage).value == expected_extended_key_usage
 
 
-def test_issue_client_certificate_has_correct_public_key():
-    ca_hierarchy = gimmecert.crypto.generate_ca_hierarchy('My Project', 1)
+@pytest.mark.parametrize("key_specification", [
+    ("rsa", 2048),
+    ("ecdsa", cryptography.hazmat.primitives.asymmetric.ec.SECP192R1)
+])
+def test_issue_client_certificate_has_correct_public_key(key_specification):
+    ca_hierarchy = gimmecert.crypto.generate_ca_hierarchy('My Project', 1, gimmecert.crypto.KeyGenerator("rsa", 2048))
     issuer_private_key, issuer_certificate = ca_hierarchy[0]
 
-    private_key = gimmecert.crypto.generate_private_key()
+    private_key = gimmecert.crypto.KeyGenerator(*key_specification)()
 
     certificate = gimmecert.crypto.issue_client_certificate('myclient', private_key.public_key(), issuer_private_key, issuer_certificate)
 
@@ -473,10 +490,10 @@ def test_issue_client_certificate_has_correct_public_key():
 
 @freeze_time('2018-01-01 00:15:00')
 def test_issue_client_certificate_not_before_is_15_minutes_in_past():
-    ca_hierarchy = gimmecert.crypto.generate_ca_hierarchy('My Project', 1)
+    ca_hierarchy = gimmecert.crypto.generate_ca_hierarchy('My Project', 1, gimmecert.crypto.KeyGenerator("rsa", 2048))
     issuer_private_key, issuer_certificate = ca_hierarchy[0]
 
-    private_key = gimmecert.crypto.generate_private_key()
+    private_key = gimmecert.crypto.KeyGenerator('rsa', 2048)()
 
     certificate = gimmecert.crypto.issue_client_certificate('myclient', private_key.public_key(), issuer_private_key, issuer_certificate)
 
@@ -485,11 +502,11 @@ def test_issue_client_certificate_not_before_is_15_minutes_in_past():
 
 def test_issue_client_certificate_not_before_does_not_exceed_ca_validity():
     with freeze_time('2018-01-01 00:15:00'):
-        ca_hierarchy = gimmecert.crypto.generate_ca_hierarchy('My Project', 1)
+        ca_hierarchy = gimmecert.crypto.generate_ca_hierarchy('My Project', 1, gimmecert.crypto.KeyGenerator("rsa", 2048))
 
     issuer_private_key, issuer_certificate = ca_hierarchy[0]
 
-    private_key = gimmecert.crypto.generate_private_key()
+    private_key = gimmecert.crypto.KeyGenerator('rsa', 2048)()
 
     with freeze_time(issuer_certificate.not_valid_before - datetime.timedelta(seconds=1)):
         certificate1 = gimmecert.crypto.issue_client_certificate('myclient', private_key.public_key(), issuer_private_key, issuer_certificate)
@@ -499,11 +516,11 @@ def test_issue_client_certificate_not_before_does_not_exceed_ca_validity():
 
 def test_issue_client_certificate_not_after_does_not_exceed_ca_validity():
     with freeze_time('2018-01-01 00:15:00'):
-        ca_hierarchy = gimmecert.crypto.generate_ca_hierarchy('My Project', 1)
+        ca_hierarchy = gimmecert.crypto.generate_ca_hierarchy('My Project', 1, gimmecert.crypto.KeyGenerator("rsa", 2048))
 
     issuer_private_key, issuer_certificate = ca_hierarchy[0]
 
-    private_key = gimmecert.crypto.generate_private_key()
+    private_key = gimmecert.crypto.KeyGenerator('rsa', 2048)()
 
     with freeze_time(issuer_certificate.not_valid_after + datetime.timedelta(seconds=1)):
         certificate1 = gimmecert.crypto.issue_client_certificate('myclient', private_key.public_key(), issuer_private_key, issuer_certificate)
@@ -512,10 +529,10 @@ def test_issue_client_certificate_not_after_does_not_exceed_ca_validity():
 
 
 def test_renew_certificate_returns_certificate():
-    ca_hierarchy = gimmecert.crypto.generate_ca_hierarchy('My Project', 1)
+    ca_hierarchy = gimmecert.crypto.generate_ca_hierarchy('My Project', 1, gimmecert.crypto.KeyGenerator("rsa", 2048))
     issuer_private_key, issuer_certificate = ca_hierarchy[0]
 
-    private_key = gimmecert.crypto.generate_private_key()
+    private_key = gimmecert.crypto.KeyGenerator('rsa', 2048)()
     old_certificate = gimmecert.crypto.issue_server_certificate('myserver', private_key.public_key(), issuer_private_key, issuer_certificate)
 
     new_certificate = gimmecert.crypto.renew_certificate(old_certificate, private_key.public_key(), issuer_private_key, issuer_certificate)
@@ -523,13 +540,17 @@ def test_renew_certificate_returns_certificate():
     assert isinstance(new_certificate, cryptography.x509.Certificate)
 
 
-def test_renew_certificate_has_correct_content():
-    ca_hierarchy = gimmecert.crypto.generate_ca_hierarchy('My Project', 1)
+@pytest.mark.parametrize("key_specification", [
+    ("rsa", 2048),
+    ("ecdsa", cryptography.hazmat.primitives.asymmetric.ec.SECP192R1)
+])
+def test_renew_certificate_has_correct_content(key_specification):
+    ca_hierarchy = gimmecert.crypto.generate_ca_hierarchy('My Project', 1, gimmecert.crypto.KeyGenerator("rsa", 2048))
     issuer_private_key, issuer_certificate = ca_hierarchy[0]
 
-    private_key = gimmecert.crypto.generate_private_key()
+    private_key = gimmecert.crypto.KeyGenerator(*key_specification)()
     old_certificate = gimmecert.crypto.issue_server_certificate('myserver', private_key.public_key(), issuer_private_key, issuer_certificate)
-    public_key = gimmecert.crypto.generate_private_key().public_key()
+    public_key = gimmecert.crypto.KeyGenerator(*key_specification)().public_key()
 
     new_certificate = gimmecert.crypto.renew_certificate(old_certificate, public_key, issuer_private_key, issuer_certificate)
 
@@ -544,10 +565,10 @@ def test_renew_certificate_not_before_is_15_minutes_in_past():
 
     # Initial server certificate.
     with freeze_time('2018-01-01 00:15:00'):
-        ca_hierarchy = gimmecert.crypto.generate_ca_hierarchy('My Project', 1)
+        ca_hierarchy = gimmecert.crypto.generate_ca_hierarchy('My Project', 1, gimmecert.crypto.KeyGenerator("rsa", 2048))
         issuer_private_key, issuer_certificate = ca_hierarchy[0]
 
-        private_key = gimmecert.crypto.generate_private_key()
+        private_key = gimmecert.crypto.KeyGenerator('rsa', 2048)()
         old_certificate = gimmecert.crypto.issue_server_certificate('myserver', private_key.public_key(), issuer_private_key, issuer_certificate)
 
     # Renew certificate.
@@ -561,10 +582,10 @@ def test_renew_certificate_not_before_does_not_exceed_ca_validity():
 
     # Initial server certificate.
     with freeze_time('2018-01-01 00:15:00'):
-        ca_hierarchy = gimmecert.crypto.generate_ca_hierarchy('My Project', 1)
+        ca_hierarchy = gimmecert.crypto.generate_ca_hierarchy('My Project', 1, gimmecert.crypto.KeyGenerator("rsa", 2048))
         issuer_private_key, issuer_certificate = ca_hierarchy[0]
 
-        private_key = gimmecert.crypto.generate_private_key()
+        private_key = gimmecert.crypto.KeyGenerator('rsa', 2048)()
         old_certificate = gimmecert.crypto.issue_server_certificate('myserver', private_key.public_key(), issuer_private_key, issuer_certificate)
 
     # Renew certificate.
@@ -578,10 +599,10 @@ def test_renew_certificate_not_after_does_not_exceed_ca_validity():
 
     # Initial server certificate.
     with freeze_time('2018-01-01 00:15:00'):
-        ca_hierarchy = gimmecert.crypto.generate_ca_hierarchy('My Project', 1)
+        ca_hierarchy = gimmecert.crypto.generate_ca_hierarchy('My Project', 1, gimmecert.crypto.KeyGenerator("rsa", 2048))
         issuer_private_key, issuer_certificate = ca_hierarchy[0]
 
-        private_key = gimmecert.crypto.generate_private_key()
+        private_key = gimmecert.crypto.KeyGenerator('rsa', 2048)()
         old_certificate = gimmecert.crypto.issue_server_certificate('myserver', private_key.public_key(), issuer_private_key, issuer_certificate)
 
     # Renew certificate.
@@ -593,7 +614,7 @@ def test_renew_certificate_not_after_does_not_exceed_ca_validity():
 
 def test_generate_csr_returns_csr_with_passed_in_dn():
 
-    private_key = gimmecert.crypto.generate_private_key()
+    private_key = gimmecert.crypto.KeyGenerator('rsa', 2048)()
     subject_dn = gimmecert.crypto.get_dn('testcsr')
 
     csr = gimmecert.crypto.generate_csr(subject_dn, private_key)
@@ -605,7 +626,7 @@ def test_generate_csr_returns_csr_with_passed_in_dn():
 
 def test_generate_csr_returns_csr_with_passed_in_name():
 
-    private_key = gimmecert.crypto.generate_private_key()
+    private_key = gimmecert.crypto.KeyGenerator('rsa', 2048)()
     name = 'testcsr'
 
     expected_subject_dn = gimmecert.crypto.get_dn('testcsr')
@@ -614,3 +635,91 @@ def test_generate_csr_returns_csr_with_passed_in_name():
 
     assert csr.public_key().public_numbers() == private_key.public_key().public_numbers()
     assert csr.subject == expected_subject_dn
+
+
+@pytest.mark.parametrize("algorithm, parameters, string_representation", [
+    ("rsa", 1024, "1024-bit RSA"),
+    ("rsa", 2048, "2048-bit RSA"),
+    ("rsa", 4096, "4096-bit RSA"),
+    ("ecdsa", cryptography.hazmat.primitives.asymmetric.ec.SECP192R1, "secp192r1 ECDSA"),
+    ("ecdsa", cryptography.hazmat.primitives.asymmetric.ec.SECP224R1, "secp224r1 ECDSA"),
+    ("ecdsa", cryptography.hazmat.primitives.asymmetric.ec.SECP256K1, "secp256k1 ECDSA"),
+    ("ecdsa", cryptography.hazmat.primitives.asymmetric.ec.SECP256R1, "secp256r1 ECDSA"),
+    ("ecdsa", cryptography.hazmat.primitives.asymmetric.ec.SECP384R1, "secp384r1 ECDSA"),
+    ("ecdsa", cryptography.hazmat.primitives.asymmetric.ec.SECP521R1, "secp521r1 ECDSA"),
+])
+def test_KeyGenerator_string_representation(algorithm, parameters, string_representation):
+
+    key_generator = gimmecert.crypto.KeyGenerator(algorithm, parameters)
+    assert str(key_generator) == string_representation
+
+
+@pytest.mark.parametrize("key_specification, key_instance_type", [
+    (("rsa", 1024), cryptography.hazmat.primitives.asymmetric.rsa.RSAPrivateKey),
+    (("rsa", 2048), cryptography.hazmat.primitives.asymmetric.rsa.RSAPrivateKey),
+    (("rsa", 4096), cryptography.hazmat.primitives.asymmetric.rsa.RSAPrivateKey),
+    (("ecdsa", cryptography.hazmat.primitives.asymmetric.ec.SECP192R1), cryptography.hazmat.primitives.asymmetric.ec.EllipticCurvePrivateKey),
+    (("ecdsa", cryptography.hazmat.primitives.asymmetric.ec.SECP224R1), cryptography.hazmat.primitives.asymmetric.ec.EllipticCurvePrivateKey),
+    (("ecdsa", cryptography.hazmat.primitives.asymmetric.ec.SECP256K1), cryptography.hazmat.primitives.asymmetric.ec.EllipticCurvePrivateKey),
+    (("ecdsa", cryptography.hazmat.primitives.asymmetric.ec.SECP256R1), cryptography.hazmat.primitives.asymmetric.ec.EllipticCurvePrivateKey),
+    (("ecdsa", cryptography.hazmat.primitives.asymmetric.ec.SECP384R1), cryptography.hazmat.primitives.asymmetric.ec.EllipticCurvePrivateKey),
+    (("ecdsa", cryptography.hazmat.primitives.asymmetric.ec.SECP521R1), cryptography.hazmat.primitives.asymmetric.ec.EllipticCurvePrivateKey),
+])
+def test_KeyGenerator_instance_returns_correct_private_key_using_passed_in_key_specification(key_specification, key_instance_type):
+
+    key_generator = gimmecert.crypto.KeyGenerator(*key_specification)
+
+    private_key = key_generator()
+    public_key = private_key.public_key()
+    public_key_specification = gimmecert.crypto.key_specification_from_public_key(public_key)
+
+    assert isinstance(private_key, key_instance_type)
+    assert public_key_specification == key_specification
+
+
+@pytest.mark.parametrize("key_specification", [
+    ("rsa", 1024),
+    ("rsa", 2048),
+    ("ecdsa", cryptography.hazmat.primitives.asymmetric.ec.SECP192R1),
+    ("ecdsa", cryptography.hazmat.primitives.asymmetric.ec.SECP224R1),
+])
+def test_generate_ca_hierarchy_generates_private_keys_using_passed_in_key_generator(key_specification):
+    base_name = "My Test"
+    depth = 3
+    key_generator = gimmecert.crypto.KeyGenerator(*key_specification)
+
+    hierarchy = gimmecert.crypto.generate_ca_hierarchy(base_name, depth, key_generator)
+
+    for private_key, _ in hierarchy:
+        public_key = private_key.public_key()
+        public_key_specification = gimmecert.crypto.key_specification_from_public_key(public_key)
+
+        assert public_key_specification == key_specification
+
+
+@pytest.mark.parametrize("specification", [
+    ("rsa", 1024),
+    ("rsa", 2048),
+    ("ecdsa", cryptography.hazmat.primitives.asymmetric.ec.SECP192R1),
+    ("ecdsa", cryptography.hazmat.primitives.asymmetric.ec.SECP224R1),
+    ("ecdsa", cryptography.hazmat.primitives.asymmetric.ec.SECP256K1),
+    ("ecdsa", cryptography.hazmat.primitives.asymmetric.ec.SECP256R1),
+    ("ecdsa", cryptography.hazmat.primitives.asymmetric.ec.SECP384R1),
+    ("ecdsa", cryptography.hazmat.primitives.asymmetric.ec.SECP521R1),
+])
+def test_key_specification_from_public_key_returns_correct_algorithm_and_parameters(specification):
+    key_generator = gimmecert.crypto.KeyGenerator(specification[0], specification[1])
+    public_key = key_generator().public_key()
+
+    algorithm, parameters = gimmecert.crypto.key_specification_from_public_key(public_key)
+
+    assert (algorithm, parameters) == specification
+
+
+def test_key_specification_raises_exception_for_invalid_public_key():
+    public_key = "not_a_public_key_instance"
+
+    with pytest.raises(ValueError) as e_info:
+        gimmecert.crypto.key_specification_from_public_key(public_key)
+
+    assert str(e_info.value) == "Unsupported public key instance passed-in: \"not_a_public_key_instance\" (<class 'str'>)"
